unit Main;

interface

uses
  windows, Forms, Classes, registry, sysutils, BDE, StdCtrls, Controls,
  KAZIP, ShFolder, ShellAPI;

type
  TMainForm = class(TForm)
    lbl_info: TLabel;
    btn_cancel: TButton;
    btn_install: TButton;
    btn_ok: TButton;
    procedure btn_install_click(Sender: TObject);
    procedure btn_cancel_click(Sender: TObject);
    procedure form_show(Sender: TObject);
    procedure form_close(Sender: TObject; var Action: TCloseAction);
  end;

var
  MainForm: TMainForm;

implementation

{$R *.dfm}

const
  datazipsize = 3411286;

function BDEInstalled: boolean;
begin
  {$I-}
  Result := (dbiInit(nil) = DBIERR_NONE)
  {$I+}
end;

// http://entwickler-forum.de/archive/index.php/t-38484.html
function GetShellFolder(const CSIDL: Integer): string;
var
  pszPath: {$IFDEF DotNet}
  StringBuilder
  {$ELSE}
  PAnsiChar
  {$ENDIF};
begin
  {$IFDEF DotNet}
  pszPath := StringBuilder.Create(MAX_PATH);
  {$ELSE}
  pszPath := AllocMem(MAX_PATH);
  {$ENDIF}
  try
    if Succeeded(SHGetFolderPath(0, CSIDL, 0, 0, pszPath)) then
      Result := pszPath;
  finally
    {$IFNDEF DotNet}
    FreeMem(pszPath);
    {$ENDIF}
  end;

  // Addition
  result := IncludeTrailingPathDelimiter(result);
end;

function GetTempDir: string;
var
  Buffer: array[0..MAX_PATH] of Char;
begin
  GetTempPath(SizeOf(Buffer) - 1, Buffer);
  Result := StrPas(Buffer);
end;

// http://www.dsdt.info/tipps/?id=426
function DelDir(Dir: String): Boolean;
var
  FileOption: TSHFileOpStruct;
begin
  ZeroMemory(@FileOption,SizeOf(FileOption));
  with FileOption do
  begin
    wFunc := FO_DELETE;
    fFlags := FOF_SILENT or FOF_NOCONFIRMATION;
    pFrom := PChar(Dir + #0);
  end;
  Result := (ShFileOperation(FileOption) = 0);
end;

procedure TMainForm.btn_install_click(Sender: TObject);
var
  Ziel: string;
  reg: tregistry;
  Fehler: boolean;
  fb, ats: TFileStream;
  KaZIP: TKaZIP;
begin
  btn_cancel.enabled := not btn_cancel.enabled;
  btn_install.enabled := not btn_install.enabled;

  fehler := false;
  try
    {$I-}

    // BDE Ordner erstellen
    Ziel := GetShellFolder(CSIDL_PROGRAM_FILES_COMMON)+'Borland Shared\BDE\';
    ForceDirectories(Ziel);

    // Dateien, die aus der EXE rausgeholt werden sollen, lschen
    if (fileexists(GetTempDir + 'BDEInst.zip')) then DeleteFile(pchar(GetTempDir + 'BDEInst.zip'));

    // Installer-EXE erst abkopieren, da sonst ein Fehler bei Win9x auftritt
    copyfile(pchar(ParamStr(0)), pchar(GetTempDir+'~BDEInst.tmp'), false);

    // Archiv und Unzip-DLL aus der EXE rausholen
    fb := TFileStream.Create(GetTempDir + '~BDEInst.tmp', fmopenread);
    try
      fb.Seek(fb.Size - datazipsize, 0);
      ats := TFileStream.Create(GetTempDir + 'BDEInst.zip', fmcreate);
      try
        ats.CopyFrom(fb, datazipsize);
      finally
        ats.Free;
      end;
    finally
      fb.free;
    end;

    // Verzeichnis erneuern
    if directoryexists(GetTempDir+'BDEInst') then DelDir(GetTempDir+'BDEInst');
    ForceDirectories(GetTempDir+'BDEInst\');

    // Extrahieren der Dateien
    KaZIP := TKaZIP.Create(self);
    try
      KaZIP.UseTempFiles := true;

      KaZIP.Open(GetTempDir+'BDEInst.zip');
      KaZIP.OverwriteAction := oaOverwriteAll;
      KaZIP.ExtractAll(Ziel);
    finally
      KaZIP.Close;
      KaZIP.Free;
    end;

    // Lschen der temporen Dateien
    if (fileexists(GetTempDir + 'BDEInst.zip')) then DeleteFile(pchar(GetTempDir + 'BDEInst.zip'));
    if (fileexists(GetTempDir + '~BDEInst.tmp')) then deletefile(pchar(GetTempDir + '~BDEInst.tmp'));

    // Lschen der eventuell schon vorhandenen Dateien
    if (fileexists(GetShellFolder(CSIDL_SYSTEM) + 'BDEADMIN.cpl')) then
      deletefile(pchar(GetShellFolder(CSIDL_SYSTEM) + 'BDEADMIN.cpl'));
    if (fileexists(GetShellFolder(CSIDL_WINDOWS) + 'BDEUninstall.exe')) then
      deletefile(pchar(GetShellFolder(CSIDL_WINDOWS) + 'BDEUninstall.exe'));
    if (fileexists(GetShellFolder(CSIDL_WINDOWS) + 'delbdeuninstall.bat')) then
      deletefile(pchar(GetShellFolder(CSIDL_WINDOWS) + 'delbdeuninstall.bat'));

    // Systemsteuerungsobjekt hinzufgen
    movefile(pchar(Ziel+'BDEADMIN.cpl'), pchar(GetShellFolder(CSIDL_SYSTEM)+'BDEADMIN.cpl'));

    // Uninstaller hinzufgen
    movefile(pchar(Ziel+'BDEUninstall.exe'), pchar(GetShellFolder(CSIDL_WINDOWS)+'BDEUninstall.exe'));
    movefile(pchar(Ziel+'delbdeuninstall.bat'), pchar(GetShellFolder(CSIDL_WINDOWS)+'delbdeuninstall.bat'));

    // Registry schreiben
    reg := tregistry.create;
    try
      reg.RootKey := HKEY_LOCAL_MACHINE;
        reg.OpenKey('Software', true);
          reg.OpenKey('Borland', true);
            reg.OpenKey('Database Engine', true);
              reg.WriteString('CONFIGFILE01', Ziel+'IDAPI32.CFG');
              reg.WriteString('DLLPATH', Ziel);
              reg.WriteString('RESOURCE', '0007');
              reg.WriteString('SaveConfig', 'WIN32');
              reg.WriteString('UseCount', '1');
              reg.OpenKey('Settings', true);
                reg.OpenKey('DRIVERS', true);
                  reg.OpenKey('DBASE', true);
                    reg.CreateKey('TABLE CREATE');
                    reg.OpenKey('INIT', true);
                      reg.WriteString('LANGDRIVER', 'DBWINWE0');
                      reg.WriteString('TYPE', 'FILE');
                      reg.WriteString('VERSION', '4.0');
                      reg.CloseKey;
        reg.OpenKey('Software', true);
          reg.OpenKey('Borland', true);
            reg.OpenKey('Database Engine', true);
              reg.OpenKey('Settings', true);
                reg.OpenKey('DRIVERS', true);
                  reg.OpenKey('DBASE', true);
                    reg.OpenKey('TABLE CREATE', true);
                      reg.WriteString('LEVEL', '7');
                      reg.WriteString('MDX BLOCK SIZE', '1024');
                      reg.WriteString('MEMO FILE BLOCK SIZE', '1024');
                      reg.CloseKey;
        reg.OpenKey('Software', true);
          reg.OpenKey('Borland', true);
            reg.OpenKey('Database Engine', true);
              reg.OpenKey('Settings', true);
                reg.OpenKey('DRIVERS', true);
                  reg.OpenKey('FOXPRO', true);
                    reg.OpenKey('TABLE CREATE', true);
                      reg.WriteString('LEVEL', '25');
                      reg.CloseKey;
        reg.OpenKey('Software', true);
          reg.OpenKey('Borland', true);
            reg.OpenKey('Database Engine', true);
              reg.OpenKey('Settings', true);
                reg.OpenKey('DRIVERS', true);
                  reg.OpenKey('FOXPRO', true);
                    reg.OpenKey('INIT', true);
                      reg.WriteString('LANGDRIVER', 'DBWINUS0');
                      reg.WriteString('TYPE', 'FILE');
                      reg.WriteString('VERSION', '4.0');
                      reg.CloseKey;
        reg.OpenKey('Software', true);
          reg.OpenKey('Borland', true);
            reg.OpenKey('Database Engine', true);
              reg.OpenKey('Settings', true);
                reg.OpenKey('DRIVERS', true);
                  reg.OpenKey('FOXPRO', true);
                    reg.OpenKey('TABLE CREATE', true);
                      reg.WriteString('LEVEL', '25');
                      reg.CloseKey;
        reg.OpenKey('Software', true);
          reg.OpenKey('Borland', true);
            reg.OpenKey('Database Engine', true);
              reg.OpenKey('Settings', true);
                reg.OpenKey('DRIVERS', true);
                  reg.OpenKey('MSACCESS', true);
                    reg.OpenKey('DB OPEN', true);
                      reg.WriteString('DATABASE NAME', 'DRIVE:/PATH/DATABASE.MDB');
                      reg.WriteString('LANGDRIVER', '');
                      reg.WriteString('OPEN MODE', 'READ/WRITE');
                      reg.WriteString('SYSTEM DATABASE', '');
                      reg.WriteString('USER NAME', '');
                      reg.CloseKey;
        reg.OpenKey('Software', true);
          reg.OpenKey('Borland', true);
            reg.OpenKey('Database Engine', true);
              reg.OpenKey('Settings', true);
                reg.OpenKey('DRIVERS', true);
                  reg.OpenKey('MSACCESS', true);
                    reg.OpenKey('INIT', true);
                      reg.WriteString('DLL32', 'IDDAO32.DLL');
                      reg.WriteString('DRIVER FLAGS', '');
                      reg.WriteString('TRACE MODE', '0');
                      reg.WriteString('TYPE', 'SERVER');
                      reg.WriteString('VERSION', '1.0');
                      reg.CloseKey;
        reg.OpenKey('Software', true);
          reg.OpenKey('Borland', true);
            reg.OpenKey('Database Engine', true);
              reg.OpenKey('Settings', true);
                reg.OpenKey('DRIVERS', true);
                  reg.OpenKey('PARADOX', true);
                    reg.OpenKey('INIT', true);
                      reg.WriteString('LANGDRIVER', 'ANSIINTL');
                      reg.WriteString('TYPE', 'FILE');
                      reg.WriteString('VERSION', '4.0');
                      reg.CloseKey;
        reg.OpenKey('Software', true);
          reg.OpenKey('Borland', true);
            reg.OpenKey('Database Engine', true);
              reg.OpenKey('Settings', true);
                reg.OpenKey('DRIVERS', true);
                  reg.OpenKey('PARADOX', true);
                    reg.OpenKey('TABLE CREATE', true);
                      reg.WriteString('BLOCK SIZE', '2048');
                      reg.WriteString('FILL FACTOR', '95');
                      reg.WriteString('LEVEL', '7');
                      reg.WriteString('STRICTINTEGRTY', 'TRUE');
                      reg.CloseKey;
        reg.OpenKey('Software', true);
          reg.OpenKey('Borland', true);
            reg.OpenKey('Database Engine', true);
              reg.OpenKey('Settings', true);
                reg.OpenKey('REPOSITORIES', true);
                reg.CloseKey;
        reg.OpenKey('Software', true);
          reg.OpenKey('Borland', true);
            reg.OpenKey('Database Engine', true);
              reg.OpenKey('Settings', true);
                reg.OpenKey('SYSTEM', true);
                  reg.OpenKey('FORMATS', true);
                    reg.openkey('DATE', true);
                      reg.WriteString('FOURDIGITYEAR', 'TRUE');
                      reg.WriteString('LEADINGZEROD', 'TRUE');
                      reg.WriteString('LEADINGZEROM', 'TRUE');
                      reg.WriteString('MODE', '1');
                      reg.WriteString('SEPARATOR', '.');
                      reg.WriteString('YEARBIASED', 'FALSE');
                      reg.CloseKey;
        reg.OpenKey('Software', true);
          reg.OpenKey('Borland', true);
            reg.OpenKey('Database Engine', true);
              reg.OpenKey('Settings', true);
                reg.OpenKey('SYSTEM', true);
                  reg.OpenKey('FORMATS', true);
                    reg.openkey('DATUM', true);
                      reg.WriteString('FOURDIGITYEAR', 'FALSE');
                      reg.WriteString('LEADINGZEROD', 'FALSE');
                      reg.WriteString('LEADINGZEROM', 'FALSE');
                      reg.WriteString('MODE', '0');
                      reg.WriteString('SEPARATOR', '/');
                      reg.WriteString('YEARBIASED', 'TRUE');
                      reg.CloseKey;
        reg.OpenKey('Software', true);
          reg.OpenKey('Borland', true);
            reg.OpenKey('Database Engine', true);
              reg.OpenKey('Settings', true);
                reg.OpenKey('SYSTEM', true);
                  reg.OpenKey('FORMATS', true);
                    reg.openkey('NUMBER', true);
                      reg.WriteString('DECIMALDIGITS', '2');
                      reg.WriteString('DECIMALSEPARATOR', ',');
                      reg.WriteString('LEADINGZERON', 'TRUE');
                      reg.WriteString('THOUSANDSEPARATOR', '.');
                      reg.CloseKey;
        reg.OpenKey('Software', true);
          reg.OpenKey('Borland', true);
            reg.OpenKey('Database Engine', true);
              reg.OpenKey('Settings', true);
                reg.OpenKey('SYSTEM', true);
                  reg.OpenKey('FORMATS', true);
                    reg.openkey('TIME', true);
                      reg.WriteString('AMSTRING', 'AM');
                      reg.WriteString('MILSECONDS', 'FALSE');
                      reg.WriteString('PMSTRING', 'PM');
                      reg.WriteString('SECONDS', 'TRUE');
                      reg.WriteString('TWELVEHOUR', 'FALSE');
                      reg.CloseKey;
        reg.OpenKey('Software', true);
          reg.OpenKey('Borland', true);
            reg.OpenKey('Database Engine', true);
              reg.OpenKey('Settings', true);
                reg.OpenKey('SYSTEM', true);
                  reg.OpenKey('INIT', true);
                    reg.WriteString('AUTO ODBC', 'FALSE');
                    reg.WriteString('DATA REPOSITORY', '');
                    reg.WriteString('DEFAULT DRIVER', 'PARADOX');
                    reg.WriteString('LANGDRIVER', 'ANSIINTL');
                    reg.WriteString('LOCAL SHARE', 'FALSE');
                    reg.WriteString('LOW MEMORY USAGE LIMIT', '32');
                    reg.WriteString('MAXBUFSIZE', '2048');
                    reg.WriteString('MAXFILEHANDLES', '48');
                    reg.WriteString('MEMSIZE', '16');
                    reg.WriteString('MINBUFSIZE', '128');
                    reg.WriteString('MTS POOLING', 'FALSE');
                    reg.WriteString('SHAREDMEMLOCATION', '');
                    reg.WriteString('SHAREDMEMSIZE', '2048');
                    reg.WriteString('SQLQRYMODE', '');
                    reg.WriteString('SYSFLAGS', '0');
                    reg.WriteString('VERSION', '4.0');
                    reg.CloseKey;
        reg.OpenKey('Software', true);
          reg.OpenKey('Microsoft', true);
            reg.OpenKey('Windows', true);
              reg.OpenKey('CurrentVersion', true);
                reg.OpenKey('Uninstall', true);
                  reg.OpenKey('Borland Database Engine', true);
                    reg.WriteString('DisplayName', 'Borland Database Engine');
                    reg.WriteString('DisplayVersion', '5.0');
                    reg.WriteString('UninstallString', GetShellFolder(CSIDL_WINDOWS) + 'BDEUninstall.exe');
                    reg.CloseKey;
    finally
      reg.Free;
    end;
    {$I+}
  except
    fehler := true;
  end;

  if ParamStr(1) = '/autoinstall' then close;

  if BDEInstalled and not fehler then
    lbl_info.caption := 'Installation erfolgreich!' + #13#13 + 'Borland Database Engine wurde erfolgreich installiert.' + #13#13 + 'Sie knnen nun Datenbankanwendungen mit Borland Database' + #13 + 'Engine nutzen.' + #13#13 + 'Eine Deinstallation ist jederzeit mglich, wird jedoch nicht empfohlen.'
  else
    lbl_info.caption := 'Fehler bei der Installation!' + #13#13 + 'Borland Database Engine konnte leider nicht auf Ihrem System' + #13 + 'installiert werden.' + #13#13 + 'Versuchen Sie gegebenenfalls den Treiber neu' + #13 + 'von www.viathinksoft.de zu beziehen oder setzen Sie sich mit' + #13 + 'dem Entwickler in Verbindung (info@daniel-marschall.de).';

  btn_cancel.Visible := not btn_cancel.Visible;
  btn_install.Visible := not btn_install.Visible;
  btn_ok.Visible := not btn_ok.Visible;
end;

procedure TMainForm.form_close(Sender: TObject; var Action: TCloseAction);
begin
  if directoryexists(GetTempDir+'BDEInst') then DelDir(GetTempDir+'BDEInst');
end;

procedure TMainForm.btn_cancel_click(Sender: TObject);
begin
  close;
end;

function BDEVersion: integer;
var
  ThisVersion: SYSVersion;
begin
  DbiGetSysVersion(ThisVersion);
  Result := ThisVersion.iVersion;
end;

procedure TMainForm.form_show(Sender: TObject);
begin
  lbl_info.caption := 'Willkommen bei der Installation von Borland Database Engine!'+#13#13+'Borland Database Engine ist ein Datenbanktreiber, der bei vielen'+#13+'Datenbankanwendungen genutzt wird.' + #13#13 + 'Information: ';

  if BDEInstalled then
  begin
    if (BDEVersion < 500) then lbl_info.Caption := lbl_info.caption + 'Sie besitzen bereits eine ltere, funktionsfhige Version.';
    if (BDEVersion = 500) then lbl_info.Caption := lbl_info.caption + 'Sie besitzen bereits die gleiche, funktionsfhige Version.';
    if (BDEVersion > 500) then lbl_info.Caption := lbl_info.caption + 'Sie besitzen bereits eine neuere, funktionsfhige Version.';
  end
  else
    lbl_info.Caption := lbl_info.caption + 'Sie besitzen noch keine funktionsfhige Version.';

  lbl_info.Caption := lbl_info.caption + #13#13 + 'Klicken Sie auf Weiter, um den Treiber zu installieren.'
end;

end.

